import os
import json
from pydoc import locate
from applipy import Application, Config
from applipy_inject import Injector
from applipy.application.application import (
    BindFunction,
    ModuleManager,
    RegisterFunction,
)
from logging import Logger, getLevelName, ERROR

try:
    import yaml
except ImportError:
    yaml = None


class LoadFromConfigModuleManager(ModuleManager):

    def set_application(self, app: Application) -> None:
        self.app = app

    def configure_all(self, bind_function: BindFunction,
                      register_function: RegisterFunction) -> None:
        config = self.injector.get(Config)

        module_names = config.get('app.modules')
        if module_names:
            for module, name in ((locate(name), name) for name in module_names):
                if module:
                    self.app.install(module)
                else:
                    self._log(ERROR, f'Could not load module `{name}`')
                    raise ImportError(name)

        super().configure_all(bind_function, register_function)

    def _log(self, level, *args, **kwargs):
        try:
            self.injector.get(Logger).log(level, *args, **kwargs)
        except Exception:
            print(f"[{getLevelName(level)}]", *args)


def load_config_from_json(config_path):
    config_file = os.path.join(config_path, f'{env.lower()}.json')
    if os.path.isfile(config_file):
        with open(config_file, 'r') as f:
            config = json.load(f)
    else:
        config = {}
    return config


def load_config_from_yaml(config_path):
    config_file = os.path.join(config_path, f'{env.lower()}.yaml')
    if yaml and os.path.isfile(config_file):
        with open(config_file, 'r') as f:
            config = yaml.load(f, Loader=yaml.Loader)
    else:
        config = {}
    return config


def load_config(config_path, env):
    config = {}
    config.update(load_config_from_yaml(config_path))
    config.update(load_config_from_json(config_path))

    protocols = []
    provider_names = Config(config, []).get('config.protocols')
    if provider_names:
        for provider, name in ((locate(name), name) for name in provider_names):
            if provider:
                print(f'Adding configuration provider `{name}`')
                protocols.append(provider())
            else:
                print(f'[ERROR] Could not load configuration provider `{name}`')
                raise ImportError(name)

    return Config(config, protocols)


def main(config_path, env):
    injector = Injector()
    module_manager = LoadFromConfigModuleManager(injector)
    app = Application(load_config(config_path, env),
                      injector=injector,
                      module_manager=module_manager)
    module_manager.set_application(app)
    app.run()


if __name__ == '__main__':
    config_path = os.environ.get('APPLIPY_CONFIG_PATH', os.path.curdir)
    env = os.environ.get('APPLIPY_ENV', 'DEV')
    main(config_path, env)



from applipy import Application, Config, AppHandle
from collections import defaultdict


class App(AppHandle):

    N = 0

    def __init__(self, call_logs: dict):
        self.call_log = call_logs[App.N]
        App.N += 1

    async def on_init(self):
        self.call_log.append('on_init')

    async def on_start(self):
        self.call_log.append('on_start')
        raise ValueError

    async def on_shutdown(self):
        self.call_log.append('on_shutdown')


def test_applipy_all_lifecyle_methods_are_called_in_the_right_order():
    app = Application(Config({}))
    call_logs = defaultdict(list)
    app.injector.bind(dict, call_logs)
    app.register(App)
    app.register(App)
    app.register(App)
    app.register(App)

    app.run()

    assert len(call_logs) == 4
    for k in call_logs:
        call_log = call_logs[k]
        assert len(call_log) == 3
        assert call_log == ['on_init', 'on_start', 'on_shutdown']



import typing as T
from .protocol import ConfigProtocol
from collections.abc import Mapping


class Config(Mapping):

    _providers: T.List[ConfigProtocol]
    _config: T.Dict[str, T.Any]
    _keys: T.Set[str]

    def __init__(self, config: T.Dict[str, T.Any], protocols: T.List[ConfigProtocol] = [],
                 *, _normalize=True) -> None:
        self._providers = protocols
        self._config = self._build_config(config) if _normalize else config
        self._keys = self._build_keys(config, [])

    def __getitem__(self, key) -> T.Optional[T.Any]:
        parts = key.split('.')

        value = self._config
        for part in parts:
            if isinstance(value, dict):
                value = value.get(part)
            else:
                raise KeyError(key)
        if value is None:
            raise KeyError(key)

        if isinstance(value, str):
            idx = value.find(':')
            scheme, other = value[:idx], value[idx + 1:]
            if len(other) > 0:
                maybe_value = self._get_value_from_provider(scheme, other)
                if maybe_value is not None:
                    value = maybe_value
        elif isinstance(value, dict):
            value = Config(value, self._providers, _normalize=False)

        return value

    def __iter__(self):
        yield from self._keys

    def __len__(self):
        return len(self._keys)

    def _build_config(self, config):
        new_config = {}
        for k, v in config.items():
            parts = k.split('.')
            sub_config = new_config
            for part in parts[:-1]:
                if part not in sub_config:
                    sub_config[part] = {}
                sub_config = sub_config[part]

            if isinstance(v, dict):
                v = self._build_config(v)

            sub_config[parts[-1]] = v

        return new_config

    def _build_keys(self, config, parts):
        keys = set()
        for k, v in config.items():
            new_parts = parts + [k]
            if isinstance(v, dict):
                keys.update(self._build_keys(v, new_parts))
            else:
                keys.add('.'.join(new_parts))

        return keys

    def _get_value_from_provider(self, protocol: str, key: str) -> T.Optional[T.Any]:
        for provider in self._providers:
            r = provider.provide_for(protocol, key)
            if r is not None:
                return r
        return None



from applipy import Module as Module_, AppHandle, LoggingModule
from asyncio import sleep
from .common import ApplipyProcess
from logging import Logger
from tempfile import NamedTemporaryFile


class App(AppHandle):

    def __init__(self, logger: Logger):
        self._logger = logger.getChild('test-app')

    async def on_init(self):
        self._logger.debug("on_init")

    async def on_start(self):
        self._logger.debug("on_start")
        while True:
            await sleep(30)

    async def on_shutdown(self):
        self._logger.debug("on_shutdown")


class Module(Module_):

    def configure(self, bind, register):
        register(App)

    @classmethod
    def depends_on(cls):
        return LoggingModule,


def test_applipy_process():
    with NamedTemporaryFile('r') as f:
        with ApplipyProcess('./tests/acceptance', 'test_basic', {'log_file': f.name}) as p:
            pass
        assert p.returncode == 0
        f.seek(0)
        assert [s for s in f.readlines()
                if s.startswith('DEBUG:test_basic.test-app:')] == ['DEBUG:test_basic.test-app:on_init\n',
                                                                   'DEBUG:test_basic.test-app:on_start\n',
                                                                   'DEBUG:test_basic.test-app:on_shutdown\n']



import os
from typing import Optional
from applipy.config.protocol import ConfigProtocol


class Environment(ConfigProtocol):

    def provide_for(self, protocol: str, key: str) -> Optional[str]:
        if protocol != 'env':
            return None

        return os.environ.get(key, None)



from applipy import Config
from applipy.config.protocol import ConfigProtocol


class Upper(ConfigProtocol):

    def provide_for(self, protocol, key):
        if protocol in ('up', 'upper'):
            return key.upper()

        return None


class Capitalize(ConfigProtocol):

    def provide_for(self, protocol, key):
        if protocol == 'cap':
            return key.capitalize()

        return None


def test_config():

    raw_config = {
            'app': {'name': 'config-test', 'priority': 40},
            'app.type': 5,
            'db.password.value': 'up:seCret',
            'db.password.is_secret': True,
            'name': {'of': {'process': 'cap:foo'}},
            'name.of.file': 'upper:unknown'
    }
    protocols = [Upper(), Capitalize()]
    config = Config(raw_config, protocols)

    assert config['app.name'] == 'config-test'
    assert config.get('app.priority') == 40
    assert config.get('app.type') == 5
    assert config.get('db.password.value') == 'SECRET'
    assert config.get('db.password.is_secret') is True
    assert config['name'].get('of.process') == 'Foo'
    assert config['name']['of.file'] == 'UNKNOWN'
    assert config.get('app.not-there', 69) == 69

    assert config.keys() == {'app.name', 'app.priority', 'app.type', 'db.password.value', 'db.password.is_secret',
                             'name.of.process', 'name.of.file'}
